@@ -1,47 +1,64 @@
+"""A processor for the tabular classification task."""
+
 from __future__ import annotations
 
+from collections.abc import Iterable
+from typing import Any
+
 from explainaboard import TaskType
 from explainaboard.analysis import feature
 from explainaboard.analysis.analyses import (
     Analysis,
     AnalysisLevel,
     BucketAnalysis,
+    CalibrationAnalysis,
     ComboCountAnalysis,
 )
 from explainaboard.analysis.feature import FeatureType
+from explainaboard.info import SysOutputInfo
 from explainaboard.metrics.accuracy import AccuracyConfig
 from explainaboard.metrics.metric import MetricConfig
 from explainaboard.processors.processor import Processor
-from explainaboard.processors.processor_registry import register_processor
 
 
-@register_processor(TaskType.tabular_classification)
-class TextClassificationProcessor(Processor):
+class TabularClassificationProcessor(Processor):
+    """A processor for the tabular classification task."""
+
     @classmethod
     def task_type(cls) -> TaskType:
+        """See Processor.task_type."""
         return TaskType.tabular_classification
 
     def default_analysis_levels(self) -> list[AnalysisLevel]:
+        """See Processor.default_analysis_levels."""
         features: dict[str, FeatureType] = {
             "true_label": feature.Value(
-                dtype="string",
+                dtype=feature.DataType.STRING,
                 description="the true label of the input",
             ),
             "predicted_label": feature.Value(
-                dtype="string",
+                dtype=feature.DataType.STRING,
                 description="the predicted label",
             ),
+            "confidence": feature.Value(
+                dtype=feature.DataType.FLOAT,
+                description="the confidence of the predicted label",
+                max_value=1.0,
+                min_value=0.0,
+                optional=True,
+            ),
         }
 
         return [
             AnalysisLevel(
-                name='example',
+                name="example",
                 features=features,
                 metric_configs=self.default_metrics(),
             )
         ]
 
     def default_analyses(self) -> list[Analysis]:
+        """See Processor.default_analyses."""
         features = self.default_analysis_levels()[0].features
         # Create analyses
         analyses: list[Analysis] = [
@@ -50,7 +67,13 @@ class TextClassificationProcessor(Processor):
                 description=features["true_label"].description,
                 feature="true_label",
                 method="discrete",
-                number=15,
+                num_buckets=15,
+            ),
+            CalibrationAnalysis(
+                level="example",
+                description="calibration analysis",
+                feature="confidence",
+                num_buckets=10,
             ),
             ComboCountAnalysis(
                 level="example",
@@ -61,8 +84,15 @@ class TextClassificationProcessor(Processor):
         analyses.extend(self.continuous_feature_analyses())
         return analyses
 
+    def _statistics_func(self, samples: Iterable[Any], sys_info: SysOutputInfo):
+        return {}
+
     @classmethod
     def default_metrics(
-        cls, level="example", source_language=None, target_language=None
-    ) -> list[MetricConfig]:
-        return [AccuracyConfig(name='Accuracy')]
+        cls,
+        level: str = "example",
+        source_language: str | None = None,
+        target_language: str | None = None,
+    ) -> dict[str, MetricConfig]:
+        """See Processor.default_metrics."""
+        return {"Accuracy": AccuracyConfig()}
